import numpy as n
import os
from time import time, asctime, localtime
from numpy.random import randn, rand
from numpy import s_, dot, tile, zeros, ones, zeros_like
from data import *
from options import *
from gpumodel import *
from math import ceil, floor
from threading import Thread
import sys
import shutil
import platform

SPARSE_DAMPING = 0.6

class RBM(GPUModel):
    MODEL_NAME = "nvrbm"
    def __init__(self, op, load_dic, device_id):
        filename_options = [('num_hid', 'h'), ('sigma', 'g'), ('add_noise', 'a'), ('sparse_cost', 'sc'),
                            ('sparse_target', 'st'), ('dp_type', 'dp')]
        GPUModel.__init__(self, RBM.MODEL_NAME, op, load_dic, device_id, filename_options)

    def init_model_lib(self):
        self.libmodel.initNVRBM(self.vis_hid, self.vis_biases, self.hid_biases,
                                self.vis_hid_inc, self.vis_bias_inc, self.hid_bias_inc,
                                self.vis_stdevs, self.vis_stdevs_inc, self.neg_data_pcd, self.hid_means,
                                self.eps_w, self.eps_vb, self.eps_hb, self.eps_sigma, self.weight_cost, self.sigma,
                                self.mom, self.sparse_cost, self.sparse_target, SPARSE_DAMPING,
                                self.minibatch_size, self.cd, self.learn_sigma, self.add_noise,
                                self.device_id)

    def init_model_state(self):
        self.num_vis = self.train_data_provider.get_data_dims()
        if not self.load_dic:
            ms = self.model_state
            ms["vis_hid"] = n.array(self.init_weight_mult * randn(self.num_vis, self.num_hid).astype(n.single), order='F')
            ms["hid_biases"] = zeros((1, self.num_hid), dtype=n.single, order='F')
            ms["vis_biases"] = zeros((1, self.num_vis), dtype=n.single, order='F')
            ms["vis_hid_inc"] = zeros_like(ms["vis_hid"])
            ms["hid_bias_inc"] = zeros_like(ms["hid_biases"])
            ms["vis_bias_inc"] = zeros_like(ms["vis_biases"])
            ms["hid_means"] = self.sparse_target * ones_like(ms["hid_biases"])
            ms["vis_stdevs"] = self.sigma * ones_like(ms["vis_biases"])
            ms["vis_stdevs_inc"] = zeros_like(ms["vis_stdevs"])
            ms["neg_data_pcd"] = n.array(self.sigma * randn(self.minibatch_size, self.num_vis).astype(n.single), order='F')
        
    def print_test_results(self):
        GPUModel.print_test_results(self)
        print "\n\t\tWeights: %.6f" % n.abs(self.vis_hid).mean()
        print "\t\tHid means: %.6f" % self.hid_means.mean(),
    
if __name__ == "__main__":
    op = GPUModel.get_default_options_parser()
    op.add_option("num-hid", "num_hid", IntegerOptionParser, "Number of hidden units", set_once=True)
    op.add_option("weight-mult", "init_weight_mult", FloatOptionParser, "Initial weight stdev", set_once=True, default=0.001)
    op.add_option("eps-w", "eps_w", FloatOptionParser, "Learning rate: weights", default=0.00001)
    op.add_option("eps-hid", "eps_hb", FloatOptionParser, "Learning rate: hidden biases", default=OptionExpression("eps_w * 10"))
    op.add_option("eps-vis", "eps_vb", FloatOptionParser, "Learning rate: visible biases", default=OptionExpression("eps_w * 10"))
    op.add_option("eps-sigma", "eps_sigma", FloatOptionParser, "Learning rate: visible standard deviations", default=0)
    op.add_option("learn-sigma", "learn_sigma", BooleanOptionParser, "Learn visible stdevs?", default=0)
    op.add_option("add-noise", "add_noise", BooleanOptionParser, "Add noise to visibles?", default=0)
    op.add_option("sigma", "sigma", FloatOptionParser, "Stdev for visible units (0=binary)")
    op.add_option("mini", "minibatch_size", IntegerOptionParser, "Minibatch size", default=128)
    op.add_option("cd", "cd", IntegerOptionParser, "CD iterations", default=1)
    op.add_option("sparse-cost", "sparse_cost", FloatOptionParser, "Sparseness: cost", default=0)
    op.add_option("sparse-target", "sparse_target", FloatOptionParser, "Sparseness: target for hidden units", default=0.05)
    op.add_option("mom", "mom", FloatOptionParser, "Momentum", default=0.9)
    op.add_option("weight-cost", "weight_cost", FloatOptionParser, "Weight decay", default=0)
    
    op.options["max_test_err"].default = 2000

    op, load_dic, device_id =  GPUModel.parse_options(op)
    model = RBM(op, load_dic, device_id)
    model.train()
